Unit PolyView;
Interface
Uses
   Objects,
   Drivers,
   Views,
   App;

Type
  PPolyView = ^TPolyView;
  TPolyView = object(TView)
    Queue   : Array [1..40] of Array [1..15] of TPoint;
    Vectors : Array[1..15] of TPoint;
    Color   : Integer;
    M,N     : Integer;
    Constructor Init(Bounds : TRect; AN,AM : Byte);
    Constructor Load(Var S: TStream);
    Procedure Store(Var S: TStream);
    Procedure Drag;
    Procedure GDraw; Virtual;
  end;
  PPolyWindow = ^TPolyWindow;
  TPolyWindow = object(TWindow)
    Constructor Init;
    Procedure Drag;
  end;

Procedure RegisterPolyView;

implementation
Const
    NColors = 13;
    Colors : Array[1..NColors] of Byte =  (0,1,9,11,2,10,14,15,3,5,13,4,12);

  RPolyView : TStreamRec = (
     ObjType: 10130;
     VmtLink: Ofs(TypeOf(TPolyView)^);
     Load:    @TPolyView.Load;
     Store:   @TPolyView.Store
  );
  RPolyWindow : TStreamRec = (
     ObjType: 10131;
     VmtLink: Ofs(TypeOf(TPolyWindow)^);
     Load:    @TPolyWindow.Load;
     Store:   @TPolyWindow.Store
  );

Constructor TPolyView.Init(Bounds : TRect; AN,AM : Byte);
Var i : Byte;
  begin
    Inherited Init(Bounds);
    GrowMode:=gfGrowHiX or gfGrowHiY;
    If AN>15 then AN:=15 else If AN<3 then AN:=3;  {p設}
    If AM>40 then AM:=40 else If AM<1 then AM:=1;  {- }
    M:=AM; N:=AN;
    FillChar(Queue,SizeOf(Queue),0); Color:=1;
    For i:=1 to N do
      begin
        Color := Random(NColors-1)+1;
        Queue[1][i].x:=Random(9999);
        Queue[1][i].y:=Random(9999);
        Vectors[i].x := Random(300) + 1;
        If Random-0.5>0 then Vectors[i].x:=-Vectors[i].x;
        Vectors[i].y := Random(300) + 1;
        If Random-0.5>0 then Vectors[i].y:=-Vectors[i].y;
      end;
  end;


Procedure TPolyView.GDraw;
Var
  R         : TRect;
  MouseHere : Boolean;

  Procedure DrawPoly(k,Color : Integer);
  Var i : Integer;
    begin
      For i:=1 to N-1 do
        grLine(Queue[k][i].x,Queue[k][i].y,
               Queue[k][i+1].x,Queue[k][i+1].y,Color);
      grLine(Queue[k][N].x,Queue[k][N].y,
             Queue[k][1].x,Queue[k][1].y,Color);
    end;
  begin
    If Not GraphAvail(MouseHere) then Exit;
    GetGraphBounds(R);
    If MouseHere then HideMouse;
    DrawPoly(1,Colors[Color]);
    DrawPoly(M,(GetColor(1) shr 4) and $F);
    If MouseHere then ShowMouse;
  end;

Constructor TPolyView.Load(Var S: TStream);
  begin
    Inherited Load(S);
    S.Read(Color,2);
    S.Read(N,2);
    S.Read(M,2);
    S.Read(Queue[1],M*15*SizeOf(TPoint));
    S.Read(Vectors,SizeOf(Vectors));
  end;
Procedure TPolyView.Store(Var S: TStream);
  begin
    Inherited Store(S);
    S.Write(Color,2);
    S.Write(N,2);
    S.Write(M,2);
    S.Write(Queue[1],M*15*SizeOf(TPoint));
    S.Write(Vectors,SizeOf(Vectors));
  end;

Procedure TPolyView.Drag;
Var
   i,x,y : integer;
  begin
    Move(Queue[1],Queue[2],(M-1)*15*SizeOf(TPoint));
    For i:=1 to N do
      begin
        x:=Queue[1][i].x + Vectors[i].x;
        y:=Queue[1][i].y + Vectors[i].y;
        If (x>9999) or (x<0) then Vectors[i].x:=-Vectors[i].x;
        If (y>9999) or (y<0) then Vectors[i].y:=-Vectors[i].y;
        Queue[1][i].x:=Queue[1][i].x + Vectors[i].x;
        Queue[1][i].y:=Queue[1][i].y + Vectors[i].y;
      end;
    If Random(1000)>900 then
      begin
        If Color<NColors then Inc(Color) else Color:=1;
      end;
    GDraw;
  end;

Constructor TPolyWindow.Init;
Var
   W : PWindow;
   R : TRect;
  begin
    R.Assign(0,0,40,12);
    Inherited Init(R,'Graph demo',0);
    Palette:=wpTextWindow;
    GetExtent(R); R.grow(-1,-1);
    Insert(New(PPolyView,Init(R,Random(15),Random(40))));
  end;

Procedure TPolyWindow.Drag;
  Procedure DragPoly( P : PPolyView); Far;
    begin
      If (TypeOf(P^)=TypeOf(TPolyView)) then P^.Drag;
    end;
  begin
    ForEach(@DragPoly);
  end;


Procedure RegisterPolyView;
begin
  RegisterType(RPolyView);
  RegisterType(RPolyWindow);
end;

end.



