{************************************************}
{                                                }
{   Turbo Vision Demo                            }
{   Copyright (c) 1992 by Borland International  }
{                                                }
{************************************************}

program TVDemo;

{$X+,S-}
{$M 16384,8192,655360}

{ Turbo Vision demo program. This program uses many of the Turbo
  Vision standard and demo units, including:

    StdDlg    - Open file browser, change directory tree.
    MsgBox    - Simple dialog to display messages.
    ColorSel  - Color customization.
    Gadgets   - Shows system time and available heap space.
    AsciiTab  - ASCII table.
    Calendar  - View a month at a time
    Calc      - Desktop calculator.
    TVGHelp   - Context sensitive help.
    MouseDlg  - Mouse options dialog.
    Puzzle    - Simple brain puzzle.
    Editors   - Text Editor object.

  And of course this program includes many standard Turbo Vision
  objects and behaviors (menubar, desktop, status line, dialog boxes,
  mouse support, window resize/move/tile/cascade).
}

uses
  Dos, Objects, Drivers, Memory, Views, Menus, Dialogs, StdDlg, HistList,
  MsgBox, App, DemoCmds, Gadgets, Puzzle, Calendar, AsciiTab, Calc,
  TVGHelp, MouseDlg, Editors, PolyView;

{ If you get a FILE NOT FOUND error when compiling this program
  from a DOS IDE, change to the \BP\EXAMPLES\DOS\TVDEMO directory
  (use File|Change dir).

  This will enable the compiler to find all of the units used by
  this program.
}

const
  HeapSize = 48 * (1024 div 16);  { Save 48k heap for main program }

  { Desktop file signature information }
  SignatureLen = 21;
  DSKSignature : string[SignatureLen] = 'TV Demo Desktop File'#26;
const

  hcNoContext            =  0;
  hcViewer               =  2;
  hcPuzzle               =  3;
  hcCalculator           =  4;
  hcCalendar             =  5;
  hcAsciiTable           =  6;
  hcSystem               =  7;
  hcSAbout               =  8;
  hcSPuzzle              =  9;
  hcSCalendar            = 10;
  hcSAsciiTable          = 11;
  hcSCalculator          = 12;
  hcFile                 = 13;
  hcEdit                 = 14;
  hcShowClip             = 15;
  hcSearch               = 16;
  hcFind                 = 17;
  hcReplace              = 18;
  hcSearchAgain          = 19;
  hcWindows              = 20;
  hcOptions              = 21;
  hcOMouse               = 22;
  hcOColors              = 23;
  hcOSaveDesktop         = 24;
  hcORestoreDesktop      = 25;
  hcFOFileOpenDBox       = 26;
  hcFOName               = 27;
  hcFOFiles              = 28;
  hcFOOpenBtn            = 29;
  hcCancelBtn            = 30;
  hcOpenBtn              = 31;
  hcFCChDirDBox          = 32;
  hcOMMouseDBox          = 33;
  hcPrevious             = 65318;

var
  ClipWindow: PEditWindow;

type

  { TTVDemo }

  PTVDemo = ^TTVDemo;
  TTVDemo = object(TApplication)
    Clock: PClockView;
    Heap: PHeapView;
    constructor Init;
    procedure FileOpen(WildCard: PathStr);
    function OpenEditor(FileName: FNameStr; Visible: Boolean): PEditWindow;
    procedure GetEvent(var Event: TEvent); virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Idle; virtual;
    procedure InitMenuBar; virtual;
    procedure InitStatusLine; virtual;
    procedure LoadDesktop(var S: TStream);
    procedure OutOfMemory; virtual;
    procedure StoreDesktop(var S: TStream);
  end;

{ CalcHelpName }

function CalcHelpName: PathStr;
var
  EXEName: PathStr;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
begin
  if Lo(DosVersion) >= 3 then EXEName := ParamStr(0)
  else EXEName := FSearch('TVGDEMO.EXE', GetEnv('PATH'));
  FSplit(EXEName, Dir, Name, Ext);
  CalcHelpName := Dir+Name+'.HLP';
end;

{ TTVDemo }
constructor TTVDemo.Init;
var
  R         : TRect;
  I         : Integer;
  FileName  : PathStr;
  Ch        : Char;
begin
  Ch:=#0;
  repeat
    If Ch<>#0 then Writeln('Error!');
    Write('Enter language selector (E)nglish/(R)ussian :');
    Readln(Ch); Ch:=UpCase(Ch);
  until (Ch in ['E','R']);
  MaxHeapSize := HeapSize;
  If not LoadMessageFile(ParamStr(0),Ch) and
     not LoadMessageFile('TVGDEMO.MSG',Ch) then
    begin
      Writeln('Fatal error: Can''t find string resounce neither');
      Writeln('             in '+ParamStr(0)+' nor in TVGDEMO.MSG');
      Writeln('             Language selector = ',Ch);
      Halt(1);
    end;
  Randomize;

  inherited Init;
  RegisterObjects;
  RegisterViews;
  RegisterMenus;
  RegisterDialogs;
  RegisterApp;
  RegisterTVGHelp;
  RegisterPuzzle;
  RegisterCalendar;
  RegisterAsciiTab;
  RegisterCalc;
  RegisterEditors;
  RegisterPolyView;

  { Initialize demo gadgets }

  GetExtent(R);
  R.A.X := R.B.X - 9; R.B.Y := R.A.Y + 1;
  Clock := New(PClockView, Init(R));
  Insert(Clock);

  GetExtent(R);
  Dec(R.B.X);
  R.A.X := R.B.X - 9; R.A.Y := R.B.Y - 1;
  Heap := New(PHeapView, Init(R));
  Insert(Heap);

  DisableCommands([cmSave, cmSaveAs, cmCut, cmCopy, cmPaste, cmClear,
    cmUndo, cmFind, cmReplace, cmSearchAgain, cmCloseAll]);
  EditorDialog := StdEditorDialog;
  ClipWindow := OpenEditor('', False);
  if ClipWindow <> nil then
  begin
    Clipboard := ClipWindow^.Editor;
    Clipboard^.CanUndo := False;
  end;

  for I := 1 to ParamCount do
  begin
    FileName := ParamStr(I);
    if FileName[Length(FileName)] = '\' then
      FileName := FileName + '*.*';
    if (Pos('?', FileName) = 0) and (Pos('*', FileName) = 0) then
      OpenEditor(FExpand(FileName), True)
    else FileOpen(FileName);
  end;
end;

function TTVDemo.OpenEditor(FileName: FNameStr; Visible: Boolean): PEditWindow;
var
  P: PView;
  R: TRect;
begin
  DeskTop^.GetExtent(R);
  P := Application^.ValidView(New(PEditWindow,
    Init(R, FileName, wnNoNumber)));
  if not Visible then P^.Hide;
  DeskTop^.Insert(P);
  OpenEditor := PEditWindow(P);
end;

procedure TTVDemo.FileOpen(WildCard: PathStr);
var
  FileName: FNameStr;
begin
  FileName := '*.*';
  if ExecuteDialog(New(PFileDialog, Init(WildCard, GetS(1),GetS(2),
    fdOpenButton + fdHelpButton, 100)), @FileName) <> cmCancel then
    OpenEditor(FileName, True);
end;

procedure TTVDemo.GetEvent(var Event: TEvent);
var
  Help: PHelpWindow;
const
  HelpInUse: Boolean = False;
Var
    Ctx : Word;
begin
  inherited GetEvent(Event);
  case Event.What of
    evCommand:
      if (Event.Command = cmHelp) and not HelpInUse then
      begin
        HelpInUse := True;
        Help := New(PHelpWindow, Init(CalcHelpName));
        Ctx:=GetHelpCtx; if Ctx=0 then Ctx:=1;
        Help^.SetData(Ctx);
        if ValidView(Help) <> nil then
          begin
            ExecView(Help);
            Dispose(Help, Done);
          end;
        ClearEvent(Event);
        HelpInUse := False;
      end;
    evMouseDown:
      if Event.Buttons <> 1 then Event.What := evNothing;
  end;
end;


procedure TTVDemo.HandleEvent(var Event: TEvent);

procedure ChangeDir;
var
  D: PChDirDialog;
begin
  D := New(PChDirDialog, Init(cdNormal + cdHelpButton, 101));
  D^.HelpCtx := hcFCChDirDBox;
  ExecuteDialog(D, nil);
end;

procedure Poly;
Var
  P: PPolyWindow;
begin
  P := New(PPolyWindow, Init);
  InsertWindow(P);
end;

procedure Puzzle;
var
  P: PPuzzleWindow;
begin
  P := New(PPuzzleWindow, Init);
  P^.HelpCtx := hcPuzzle;
  InsertWindow(P);
end;

procedure Calendar;
var
  P: PCalendarWindow;
begin
  P := New(PCalendarWindow, Init);
  P^.HelpCtx := hcCalendar;
  InsertWindow(P);
end;

procedure About;
var
  D: PDialog;
  Control: PView;
  R: TRect;
begin
  R.Assign(0, 0, 40, 11);
  D := New(PDialog, Init(R, 'About'));
  with D^ do
  begin
    Options := Options or ofCentered;
    Palette := wpMessageWindow;

    R.Grow(-1, -1);
    Dec(R.B.Y, 3);
    Insert(New(PStaticText, Init(R,
      #13#3'Turbo Vision Demo'#13 +
      #3'(c) 1992 Borland International'#13+
      #13#3'Graphics:' +
      #13#3'(c) 1994 AE (Alexander Eltsyn)')));

    R.Assign(15, 8, 25, 10);
    Insert(New(PButton, Init(R, 'O~K', cmOk, bfDefault)));
  end;
  if ValidView(D) <> nil then
  begin
    Desktop^.ExecView(D);
    Dispose(D, Done);
  end;
end;

procedure AsciiTab;
var
  P: PAsciiChart;
begin
  P := New(PAsciiChart, Init);
  P^.HelpCtx := hcAsciiTable;
  InsertWindow(P);
end;

procedure Calculator;
var
  P: PCalculator;
begin
  P := New(PCalculator, Init);
  P^.HelpCtx := hcCalculator;
  InsertWindow(P);
end;

procedure Mouse;
var
  D: PDialog;
begin
  D := New(PMouseDialog, Init);
  D^.HelpCtx := hcOMMouseDBox;
  ExecuteDialog(D, @MouseReverse);
end;

procedure RetrieveDesktop;
var
  S: PStream;
  Signature: string[SignatureLen];
begin
  S := New(PBufStream, Init('TVDEMO.DSK', stOpenRead, 1024));
  if LowMemory then OutOfMemory
  else if S^.Status <> stOk then
    MessageBox(GetS(3), nil, mfOkButton + mfError)
  else
  begin
    Signature[0] := Char(SignatureLen);
    S^.Read(Signature[1], SignatureLen);
    if Signature = DSKSignature then
    begin
      LoadDesktop(S^);
      LoadHistory(S^);
      if S^.Status <> stOk then
        MessageBox(GetS(4), nil, mfOkButton + mfError);
    end
    else
      MessageBox(GetS(4), nil, mfOkButton + mfError);
  end;
  Dispose(S, Done);
end;

procedure SaveDesktop;
var
  S: PStream;
  F: File;
begin
  S := New(PBufStream, Init('TVDEMO.DSK', stCreate, 1024));
  if not LowMemory and (S^.Status = stOk) then
  begin
    S^.Write(DSKSignature[1], SignatureLen);
    StoreDesktop(S^);
    StoreHistory(S^);
    if S^.Status <> stOk then
    begin
      MessageBox(GetS(5), nil, mfOkButton + mfError);
      {$I-}
      Dispose(S, Done);
      Assign(F, 'TVDEMO.DSK');
      Erase(F);
      Exit;
    end;
  end;
  Dispose(S, Done);
end;

procedure FileNew;
begin
  OpenEditor('', True);
end;

procedure ShowClip;
begin
  ClipWindow^.Select;
  ClipWindow^.Show;
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
      begin
        case Event.Command of
          cmOpen: FileOpen('*.*');
          cmNew: FileNew;
          cmShowClip: ShowClip;
          cmChangeDir: ChangeDir;
          cmAbout: About;
          cmPoly: Poly;
          cmPuzzle: Puzzle;
          cmCalendar: Calendar;
          cmAsciiTab: AsciiTab;
          cmCalculator: Calculator;
          cmVideo: begin
                      If (ScreenMode and smVGA = smVGA)
                        then SetScreenMode(smEGA)
                        else SetScreenMode(smVGA);
                    end;
          cmMouse: Mouse;
          cmSaveDesktop: SaveDesktop;
          cmRetrieveDesktop: RetrieveDesktop;
        else
          Exit;
        end;
        ClearEvent(Event);
      end;
  end;
end;

procedure TTVDemo.Idle;
  Procedure DragPolys(P : PPolyWindow); Far;
    begin
      If (TypeOf(P^)=TypeOf(TPolyWindow)) then P^.Drag;
    end;

function IsTileable(P: PView): Boolean; far;
begin
  IsTileable := (P^.Options and ofTileable <> 0) and
    (P^.State and sfVisible <> 0);
end;

begin
  inherited Idle;
  Clock^.Update;
  Heap^.Update;
  Desktop^.ForEach(@DragPolys);
  if Desktop^.FirstThat(@IsTileable) <> nil then
    EnableCommands([cmTile, cmCascade])
  else
    DisableCommands([cmTile, cmCascade]);
end;

procedure TTVDemo.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y+1;
  MenuBar := New(PMenuBar, Init(R, NewMenu(
    NewSubMenu('~'#240'~', hcSystem, NewMenu(
      NewItem(GetS(10), '', kbNoKey, cmAbout, hcSAbout,
      NewLine(
      NewItem('Graph poly ', '', kbNoKey, cmPoly, hcCalculator,
      NewItem(GetS(11), '', kbNoKey, cmPuzzle, hcSPuzzle,
      NewItem(GetS(12), '', kbNoKey, cmCalendar, hcSCalendar,
      NewItem(GetS(13), '', kbNoKey, cmAsciiTab, hcSAsciiTable,
      NewItem(GetS(14), '', kbNoKey, cmCalculator, hcCalculator, nil)))))))),
    NewSubMenu(GetS(15), hcFile, NewMenu(
      StdFileMenuItems(nil)),
    NewSubMenu(GetS(16), hcEdit, NewMenu(
      StdEditMenuItems(
      NewLine(
      NewItem(GetS(17), '', kbNoKey, cmShowClip, hcShowClip,
      nil)))),
    NewSubMenu(GetS(20), hcSearch, NewMenu(
      NewItem(GetS(21), '', kbNoKey, cmFind, hcFind,
      NewItem(GetS(22), '', kbNoKey, cmReplace, hcReplace,
      NewItem(GetS(23), '', kbNoKey, cmSearchAgain, hcSearchAgain,
      nil)))),
    NewSubMenu(GetS(30), hcWindows, NewMenu(
      StdWindowMenuItems(nil)),
    NewSubMenu(GetS(40), hcOptions, NewMenu(
      NewItem(GetS(41), '', kbNoKey, cmMouse, hcOMouse,
      NewItem(GetS(42), '', kbNoKey, cmVideo, hcNoContext,
      NewLine(
      NewItem(GetS(43), '', kbNoKey, cmSaveDesktop, hcOSaveDesktop,
      NewItem(GetS(44), '', kbNoKey, cmRetrieveDesktop, hcORestoreDesktop, nil)))))),
      nil)))))))));
end;

procedure TTVDemo.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  StatusLine := New(PStatusLine, Init(R,
    NewStatusDef(0, $FFFF,
      NewStatusKey(GetS(50), kbAltX, cmQuit,
      NewStatusKey(GetS(51), kbF1, cmHelp,
      NewStatusKey(GetS(52), kbF3, cmOpen,
      NewStatusKey(GetS(53), kbAltF3, cmClose,
      NewStatusKey(GetS(54), kbF5, cmZoom,
      NewStatusKey('', kbF10, cmMenu,
      NewStatusKey('', kbCtrlF5, cmResize,
      nil))))))),
    nil)));
end;

procedure TTVDemo.OutOfMemory;
begin
  MessageBox(GetS(60), nil, mfError + mfOkButton);
end;

{ Since the safety pool is only large enough to guarantee that allocating
  a window will not run out of memory, loading the entire desktop without
  checking LowMemory could cause a heap error.  This means that each
  window should be read individually, instead of using Desktop's Load.
}

procedure TTVDemo.LoadDesktop(var S: TStream);
var
  P: PView;
  Pal: PString;

procedure CloseView(P: PView); far;
begin
  Message(P, evCommand, cmClose, nil);
end;

begin
  if Desktop^.Valid(cmClose) then
  begin
    Desktop^.ForEach(@CloseView); { Clear the desktop }
    repeat
      P := PView(S.Get);
      Desktop^.InsertBefore(ValidView(P), Desktop^.Last);
    until P = nil;
    Pal := S.ReadStr;
    if Pal <> nil then
    begin
      Application^.GetPalette^ := Pal^;
      DoneMemory;
      Application^.ReDraw;
      DisposeStr(Pal);
    end;
  end;
end;

procedure TTVDemo.StoreDesktop(var S: TStream);
var
  Pal: PString;

procedure WriteView(P: PView); far;
begin
  if P <> Desktop^.Last then S.Put(P);
end;

begin
  Desktop^.ForEach(@WriteView);
  S.Put(nil);
  Pal := @Application^.GetPalette^;
  S.WriteStr(Pal);
end;


var
  Demo: TTVDemo;

begin
  Demo.Init;
  Demo.Run;
  Demo.Done;
end.
