{*******************************************************}
{                                                       }
{ File name         : VLDS.PAS                          }
{ Last edition date : 01-Apr-1994,10:49:46              }
{                                                       }
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Turbo Vision Unit                               }
{                                                       }
{       Copyright (c) 1994 AE (A. Eltsyn)               }
{                                                       }
{*******************************************************}
Unit
    Vlds;
Interface
Uses
    Objects,
    Drivers,
    MSGBOX,
    Validate;

Type
  PIntValidator = ^TIntValidator;
  TIntValidator = Object(TFilterValidator)
                     min,max  : double;
                     Constructor Init(minv,maxv : Integer);
                     procedure Error;                            Virtual;
                     function IsValid(const S: string): Boolean; Virtual;
                     function Transfer(Var S      : String;
                                           Buffer : Pointer;
                                           Flag   : TVTransfer
                                      ) : Word;                  Virtual;
                     Constructor Load(var S: TStream);
                     Procedure Store(var S: TStream);            Virtual;
                   end;

  PLIntValidator = ^TLIntValidator;
  TLIntValidator = Object(TIntValidator)
                     Constructor Init(minv,maxv : longint);
                     function IsValid(const S: string): Boolean; Virtual;
                     function Transfer(Var S      : String;
                                           Buffer : Pointer;
                                           Flag   : TVTransfer
                                      ) : Word;                  Virtual;
                   end;

  PRealValidator = ^TRealValidator;
  TRealValidator = Object(TFilterValidator)
                     min,max  : double;
                     Constructor Init(minv,maxv : Real);
                     procedure Error;                            Virtual;
                     function IsValid(const S: string): Boolean; Virtual;
                     function Transfer(Var S      : String;
                                           Buffer : Pointer;
                                           Flag   : TVTransfer
                                      ) : Word;                  Virtual;
                     Constructor Load(Var S: TStream);
                     Procedure Store(Var S: TStream);
                   end;

Procedure RegisterVlds;

Implementation
Const
  RIntValidator: TStreamRec = ( ObjType: 1010 ;
                                VmtLink: Ofs(TypeOf(TIntValidator)^);
                                Load: @TIntValidator.Load;
                                Store: @TIntValidator.Store);
  RLIntValidator: TStreamRec = ( ObjType: 1011 ;
                                VmtLink: Ofs(TypeOf(TLIntValidator)^);
                                Load: @TLIntValidator.Load;
                                Store: @TLIntValidator.Store);

  RRealValidator: TStreamRec = ( ObjType: 1012 ;
                                VmtLink: Ofs(TypeOf(TRealValidator)^);
                                Load: @TRealValidator.Load;
                                Store: @TRealValidator.Store);

{}
Constructor TIntValidator.Init(minv,maxv : integer);
  begin
    Inherited Init(['0'..'9','-']);
    Options:=Options or voTransfer;
    min:=minv;max:=maxv;
  end;
{}
Procedure TIntValidator.Error;
Var l : Array[0..1] of LongInt;
  begin
    l[0]:=Round(Min); l[1]:=Round(Max);
    MessageBox(GetS(60700), @l, mfError or mfCancelButton);
  end;
{}
Function TIntValidator.IsValid(const S : string) : Boolean;
Var i : Integer;
    r : Integer;
  begin
    Val(s,r,i);
    IsValid:=(s<>'') AND (i=0) AND (r<=max) AND (r>=min);
  end;
{}
function TIntValidator.Transfer(Var S      : String;
                                     Buffer : Pointer;
                                     Flag   : TVTransfer
                                ) : Word;
Var i : Integer;
    r : Integer;
  begin
    Transfer:=2;
    Case Flag of
      vtDataSize:;
      vtGetData:begin
                  Val(s,r,i);
                  Move(r,Buffer^,2);
                end;
      vtSetData:begin
                  Move(Buffer^,r,2);
                  Str(r:1,s);
                end;
    end;
  end;
{}
Constructor TIntValidator.Load(Var S: TStream);
  begin
    Inherited Load(S);
    S.Read(min,8);
    S.Read(max,8);
  end;
{}
Procedure TIntValidator.Store(Var S: TStream);
  begin
    Inherited Store(S);
    S.Write(min,8);
    S.Write(max,8);
  end;
{}
Constructor TLIntValidator.Init(minv,maxv : LongInt);
  begin
    TFilterValidator.Init(['0'..'9','-']);
    Options:=Options or voTransfer;
    min:=minv;max:=maxv;
  end;
{}
Function TLIntValidator.IsValid(const S : string) : Boolean;
Var i : Integer;
    r : LongInt;
  begin
    Val(s,r,i);
    IsValid:=(s<>'') AND (i=0) AND (r<=max) AND (r>=min);
  end;
{}
function TLIntValidator.Transfer(Var S      : String;
                                     Buffer : Pointer;
                                     Flag   : TVTransfer
                                ) : Word;
Var i : Integer;
    r : LongInt;
  begin
    Transfer:=2;
    Case Flag of
      vtDataSize:;
      vtGetData:begin
                  Val(s,r,i);
                  Move(r,Buffer^,4);
                end;
      vtSetData:begin
                  Move(Buffer^,r,4);
                  Str(r,s);
                end;
    end;
  end;
{}
Constructor TRealValidator.Init(minv,maxv : real);
  begin
    Inherited Init(['0'..'9','.','-']);
    Options:=Options or voTransfer;
    min:=minv; max:=maxv;
  end;
{}
Procedure TRealValidator.Error;
Var l : Array[0..1] of String;
  begin
    Str(Min:1:3,l[0]);
    Str(Max:1:3,l[1]);
    MessageBox(#3' ᫮  室  '#13#10#3+
               l[0]+'..'+l[1], Nil, mfError or mfCancelButton);
  end;
{}
Function TRealValidator.IsValid(const S : string) : Boolean;
Var i : Integer;
    r : Real;
  begin
    Val(s,r,i);
    IsValid:=(s<>'') AND (i=0) AND ((max-r)>0) AND ((r-min)>0);
  end;
{}
function TRealValidator.Transfer(Var S      : String;
                                     Buffer : Pointer;
                                     Flag   : TVTransfer
                                ) : Word;
Var i : Integer;
    r : Real;
  begin
    Transfer:=SizeOf(Real);
    Case Flag of
      vtDataSize:;
      vtGetData : begin
                    Val(s,r,i);
                    Move(r,Buffer^,SizeOf(r));
                  end;
      vtSetData : begin
                    Move(Buffer^,r,SizeOf(r));
                    Str(r:5:4,s);
                  end;
    end;
  end;
{}
{}
Constructor TRealValidator.Load(Var S: TStream);
  begin
    Inherited Load(S);
    S.Read(min,SizeOf(min));
    S.Read(max,SizeOf(max));
  end;
{}
Procedure TRealValidator.Store(Var S: TStream);
  begin
    Inherited Store(S);
    S.Write(min,SizeOf(min));
    S.Write(max,SizeOf(max));
  end;
{}
{}
Procedure RegisterVlds;
  begin
    RegisterType(RIntValidator);
    RegisterType(RLIntValidator);
    RegisterType(RRealValidator);
  end;

end.
